import customtkinter
from .main_window import MainWindow


def launch_GUI ():
    customtkinter.set_appearance_mode("System")
    customtkinter.set_default_color_theme("dark-blue")
    app = customtkinter.CTk()
    MainWindow(app)
    app.mainloop()


if __name__ == "__main__":
    launch_GUI ()


import os
import csv
import json
import xml.etree.ElementTree as ET
from xml.dom import minidom


def save_yolo_format(detections, labels_dir, image_file):
    result_txt_path = os.path.join(labels_dir, f"{os.path.splitext(image_file)[0]}.txt")
    with open(result_txt_path, "w") as txt_file:
        for (
            cls_id,
            bbox_x,
            bbox_y,
            bbox_width,
            bbox_height,
            _,
            _,
            img_width,
            img_height,
        ) in detections:
            normalized_x = bbox_x / img_width
            normalized_y = bbox_y / img_height
            normalized_w = bbox_width / img_width
            normalized_h = bbox_height / img_height
            txt_file.write(f"{cls_id} {normalized_x:.6f} {normalized_y:.6f} {normalized_w:.6f} {normalized_h:.6f}\n")


def save_coco_format(detections, image_file, i, coco_output_file):
    if not os.path.isfile(coco_output_file):
        coco_output = {
            "info": {},
            "licenses": [],
            "images": [],
            "annotations": [],
            "categories": [],
        }
    else:
        with open(coco_output_file, "r") as f:
            coco_output = json.load(f)

    for (
        cls_id,
        bbox_x,
        bbox_y,
        bbox_width,
        bbox_height,
        _,
        _,
        img_width,
        img_height,
    ) in detections:
        coco_output["images"].append(
            {"id": i, "file_name": image_file, "height": img_height, "width": img_width}
        )
        annotation_id = len(coco_output["annotations"]) + 1

        bbox_xmin = int(bbox_x - bbox_width / 2)
        bbox_ymin = int(bbox_y - bbox_height / 2)
        coco_output["annotations"].append(
            {
                "id": annotation_id,
                "image_id": i,
                "category_id": int(cls_id),
                "bbox": [bbox_xmin, bbox_ymin, int(bbox_width), int(bbox_height)],
                "area": int(bbox_width * bbox_height),
                "iscrowd": 0,
            }
        )

    with open(coco_output_file, "w") as f:
        json.dump(coco_output, f)


def save_csv_format(detections, image_file, csv_file_path, csv_headers):
    file_exists = os.path.isfile(csv_file_path)
    with open(csv_file_path, "a", newline="") as csv_file:
        csv_writer = csv.writer(csv_file)
        if not file_exists:
            csv_writer.writerow(csv_headers)
        for (
            _,
            bbox_x,
            bbox_y,
            bbox_width,
            bbox_height,
            label_name,
            _,
            img_width,
            img_height,
        ) in detections:
            row_data = [
                label_name,
                float(bbox_x - bbox_width / 2),
                float(bbox_y - bbox_height / 2),
                float(bbox_width),
                float(bbox_height),
                image_file,
                int(img_width),
                int(img_height),
            ]
            csv_writer.writerow(row_data)


def save_xml_format(detections, labels_dir, image_file):
    result_xml_path = os.path.join(labels_dir, f"{os.path.splitext(image_file)[0]}.xml")
    root = ET.Element("annotation")

    ET.SubElement(root, "folder").text = os.path.basename(labels_dir)
    ET.SubElement(root, "filename").text = image_file
    ET.SubElement(root, "path").text = os.path.join(labels_dir, image_file)

    source = ET.SubElement(root, "source")
    ET.SubElement(source, "database").text = "Unspecified"

    for (
        _,
        bbox_x,
        bbox_y,
        bbox_width,
        bbox_height,
        label_name,
        _,
        img_width,
        img_height,
    ) in detections:
        size = ET.SubElement(root, "size")
        ET.SubElement(size, "width").text = str(img_width)
        ET.SubElement(size, "height").text = str(img_height)
        ET.SubElement(size, "depth").text = "3"

        obj = ET.SubElement(root, "object")
        ET.SubElement(obj, "name").text = str(label_name)
        ET.SubElement(obj, "pose").text = "Unspecified"
        ET.SubElement(obj, "truncated").text = "0"
        ET.SubElement(obj, "difficult").text = "0"

        bndbox = ET.SubElement(obj, "bndbox")
        ET.SubElement(bndbox, "xmin").text = str(int(bbox_x - bbox_width / 2))
        ET.SubElement(bndbox, "ymin").text = str(int(bbox_y - bbox_height / 2))
        ET.SubElement(bndbox, "xmax").text = str(int(bbox_x + bbox_width / 2))
        ET.SubElement(bndbox, "ymax").text = str(int(bbox_y + bbox_height / 2))

    xml_str = ET.tostring(root, encoding="utf-8", method="xml")
    pretty_xml_str = minidom.parseString(xml_str).toprettyxml(indent="  ")
    pretty_xml_str = "\n".join(pretty_xml_str.split("\n")[1:])

    with open(result_xml_path, "w") as f:
        f.write(pretty_xml_str)



import threading
from queue import Queue
from typing import Union
from .utils.chart_utils import (
    input_chart,
    output_charts,
    close_matplotlib_figures,
    display_chart,
)


class ChartWorker(threading.Thread):
    def __init__(
        self,
        images_folder,
        chart_colors,
        train_detections=None,
        val_detections=None,
        actual_train_count=0,
        actual_val_count=0,
    ):
        super().__init__()
        self.images_folder = images_folder
        self.chart_colors = chart_colors
        self.train_detections = train_detections if train_detections is not None else []
        self.val_detections = val_detections if val_detections is not None else []
        self.actual_train_count = actual_train_count
        self.actual_val_count = actual_val_count
        self.input_figures = None
        self.train_figures = None
        self.val_figures = None
        self.image_files = []
        self.queue: Union[Queue, None] = None

    def run(self):
        self.process_input_charts()
        self.process_output_charts()

    def process_input_charts(self):
        try:
            (
                fig_format,
                fig_resolution,
                fig_quality,
                fig_colorspace,
                self.image_files,
            ) = input_chart(self.images_folder, self.chart_colors)
            self.input_figures = [
                fig_format,
                fig_resolution,
                fig_quality,
                fig_colorspace,
            ]
            if self.queue is not None:
                self.queue.put(("input_figures", self.input_figures))
        except Exception as e:
            if self.queue is not None:
                self.queue.put(("error", str(e)))

    def process_output_charts(self):
        if self.train_detections:
            train_fig_count, train_fig_conf, train_fig_labels, train_fig_heatmap = (
                output_charts(self.train_detections, self.actual_train_count)
            )
            self.train_figures = [
                train_fig_count,
                train_fig_conf,
                train_fig_labels,
                train_fig_heatmap,
            ]
            if self.queue is not None:
                self.queue.put(("train_figures", self.train_figures))
        if self.val_detections:
            val_fig_count, val_fig_conf, val_fig_labels, val_fig_heatmap = (
                output_charts(self.val_detections, self.actual_val_count)
            )
            self.val_figures = [
                val_fig_count,
                val_fig_conf,
                val_fig_labels,
                val_fig_heatmap,
            ]
            if self.queue is not None:
                self.queue.put(("val_figures", self.val_figures))

    def get_input_figures(self):
        return self.input_figures

    def get_train_figures(self):
        return self.train_figures

    def get_val_figures(self):
        return self.val_figures

    def close_figures(self):
        close_matplotlib_figures()

    def display_chart_in_frame(self, fig, frame, row, col):
        display_chart(fig, frame, row, col)



import cv2 as cv
import random

def draw_boxes_on_image(annotated_frame, detections, class_colors, output_path):
    for (
        cls_id,
        x_scaled,
        y_scaled,
        w_scaled,
        h_scaled,
        class_name,
        confidence,
        *_,
    ) in detections:
        x1 = int(x_scaled - w_scaled / 2)
        y1 = int(y_scaled - h_scaled / 2)
        x2 = int(x_scaled + w_scaled / 2)
        y2 = int(y_scaled + h_scaled / 2)

        if cls_id not in class_colors:
            class_colors[cls_id] = (
                random.randint(0, 255),
                random.randint(0, 255),
                random.randint(0, 255),
            )

        color = class_colors[cls_id]

        img_height, img_width = annotated_frame.shape[:2]
        
        object_width = x2 - x1
        object_height = y2 - y1
        size_ratio = min(object_width, object_height) / min(img_width, img_height)
        font_scale = min(0.8, 0.4 + size_ratio * 0.6)
        text_thickness = max(1, int(font_scale * 2))

        cv.rectangle(annotated_frame, (x1, y1), (x2, y2), color, 2)

        caption = f"{class_name} {confidence*100:.2f}%"
        (tw, th), _ = cv.getTextSize(
            text=caption,
            fontFace=cv.FONT_HERSHEY_SIMPLEX,
            fontScale=font_scale,
            thickness=text_thickness,
        )
        th = int(th * 1.2)
        tw = int(tw * 1.1)

        bg_x1 = max(x1, 0)
        bg_y1 = max(y1 - th, 0)
        bg_x2 = min(x1 + tw, img_width)
        bg_y2 = y1

        cv.rectangle(annotated_frame, (bg_x1, bg_y1), (bg_x2, bg_y2), color, -1)
        cv.putText(
            annotated_frame,
            caption,
            (x1, y1),
            cv.FONT_HERSHEY_SIMPLEX,
            font_scale,
            (255, 255, 255),
            text_thickness,
            cv.LINE_AA,
        )

    cv.imwrite(output_path, annotated_frame)
    return annotated_frame


import threading
import cv2 as cv
import os
from queue import Queue
from typing import Union
from ultralytics import YOLOWorld
from ultralytics.utils.torch_utils import select_device
from .utils.detection_utils import infer, slice_and_infer
from .utils.draw_utils import draw_boxes_on_image
from .utils.write_utils import (
    save_yolo_format,
    save_coco_format,
    save_csv_format,
    save_xml_format,
)


class ObjectDetectionWorker(threading.Thread):
    def __init__(
        self,
        conf_threshold,
        model_path,
        images_folder,
        output_directory,
        val_split,
        output_format,
        classes,
        device,
        use_sahi,
        slice_size,
        iou_threshold,
    ):
        super().__init__()
        self.conf_threshold = conf_threshold
        self.iou_threshold = iou_threshold
        self.model_path = model_path
        self.images_folder = images_folder
        self.output_directory = output_directory
        self.val_split = val_split
        self.output_format = output_format
        self.classes = classes
        self.device = device
        self.use_sahi = use_sahi
        self.slice_size = slice_size
        self._terminate_event = threading.Event()
        self.train_detections = []
        self.val_detections = []
        self.class_colors = {}

        self.log_queue: Union[Queue, None] = None
        self.output_image_queue: Union[Queue, None] = None
        self.progress_queue: Union[Queue, None] = None
        self.show_error_queue: Union[Queue, None] = None

    def get_image_files(self):
        return [
            f
            for f in os.listdir(self.images_folder)
            if f.endswith((".jpg", ".jpeg", ".png", "jfif"))
        ]

    def run(self):
        select_device(self.device, verbose=False)
        image_files = self.get_image_files()
        if not image_files:
            self.log_update("No images in the folder. Exiting.")
            self.show_error("No images found in the selected folder.")
            return

        num_total_images = len(image_files)
        num_val_images = int(num_total_images * self.val_split)
        num_train_images = num_total_images - num_val_images

        self.actual_train_count = 0
        self.actual_val_count = 0

        self.create_output_directories()

        model = self.load_model()

        for i, image_file in enumerate(image_files):
            if self._terminate_event.is_set():
                break

            img = self.load_image(image_file)
            if img is None:
                continue

            detections, annotated_frame = self.perform_inference(model, img, image_file)

            has_detections = len(detections) > 0
            (
                output_dir,
                detected_image_dir,
                labels_dir,
                original_dir,
                detections_list,
            ) = self.get_output_paths(i, num_val_images, has_detections)
            detections_list.extend([(det, image_file) for det in detections])

            self.save_results(
                detections,
                annotated_frame,
                img,
                image_file,
                detected_image_dir,
                labels_dir,
                original_dir,
                i,
            )
            self.progress(int((i + 1) / num_total_images * 100))
        self.finalize_detections(num_train_images, num_val_images, num_total_images)

    def get_output_paths(self, i, num_val_images, has_detections):
        if i < num_val_images:
            output_dir = os.path.join(self.output_directory, "val")
            detected_image_dir = os.path.join(output_dir, "detected")
            labels_dir = os.path.join(output_dir, "labels")
            original_dir = os.path.join(output_dir, "images")
            detections_list = self.val_detections
            if has_detections:
                self.actual_val_count += 1
        else:
            output_dir = os.path.join(self.output_directory, "train")
            detected_image_dir = os.path.join(output_dir, "detected")
            labels_dir = os.path.join(output_dir, "labels")
            original_dir = os.path.join(output_dir, "images")
            detections_list = self.train_detections
            if has_detections:
                self.actual_train_count += 1

        return output_dir, detected_image_dir, labels_dir, original_dir, detections_list

    def create_output_directories(self):
        if self.val_split < 1:
            train_dir = os.path.join(self.output_directory, "train")
            detected_image_dir = os.path.join(train_dir, "detected")
            train_labels_dir = os.path.join(train_dir, "labels")
            train_original_dir = os.path.join(train_dir, "images")
            os.makedirs(detected_image_dir, exist_ok=True)
            os.makedirs(train_labels_dir, exist_ok=True)
            os.makedirs(train_original_dir, exist_ok=True)

        if self.val_split > 0:
            val_dir = os.path.join(self.output_directory, "val")
            detected_image_dir = os.path.join(val_dir, "detected")
            val_labels_dir = os.path.join(val_dir, "labels")
            val_original_dir = os.path.join(val_dir, "images")
            os.makedirs(detected_image_dir, exist_ok=True)
            os.makedirs(val_labels_dir, exist_ok=True)
            os.makedirs(val_original_dir, exist_ok=True)

    def load_model(self):
        model = YOLOWorld(self.model_path)
        model.set_classes(self.classes)
        return model

    def load_image(self, image_file):
        img_path = os.path.join(self.images_folder, image_file)
        return cv.imread(img_path)

    def perform_inference(self, model, img, image_file):
        if self.use_sahi:
            return slice_and_infer(
                model,
                img,
                self.conf_threshold,
                self.log_update,
                image_file,
                self.slice_size,
                self.iou_threshold,
                self.terminate,
            )
        else:
            return infer(
                model,
                img,
                self.conf_threshold,
                self.log_update,
                image_file,
                self.iou_threshold,
            )

    def save_results(
        self,
        detections,
        annotated_frame,
        img,
        image_file,
        detected_image_dir,
        labels_dir,
        original_dir,
        i,
    ):
        if img is None:
            return

        if detections:
            output_img = draw_boxes_on_image(
                annotated_frame,
                detections,
                self.class_colors,
                os.path.join(detected_image_dir, image_file),
            )

            output_img_path = os.path.join(detected_image_dir, image_file)
            cv.imwrite(output_img_path, output_img)

            self.save_format_results(detections, labels_dir, image_file, i)

            original_img_path = os.path.join(original_dir, image_file)
            cv.imwrite(original_img_path, img)

            if self.output_image_queue is not None:
                self.output_image(output_img_path, image_file)

    def save_format_results(self, detections, labels_dir, image_file, i):
        if self.output_format == "YOLO":
            save_yolo_format(detections, labels_dir, image_file)
        elif self.output_format == "COCO":
            coco_output_file = os.path.join(labels_dir, "coco_output.json")
            save_coco_format(detections, image_file, i, coco_output_file)
        elif self.output_format == "CSV":
            csv_headers = [
                "label_name",
                "bbox_x",
                "bbox_y",
                "bbox_width",
                "bbox_height",
                "image_name",
                "image_width",
                "image_height",
            ]
            csv_file_path = os.path.join(labels_dir, "annotations.csv")
            save_csv_format(detections, image_file, csv_file_path, csv_headers)
        elif self.output_format == "XML":
            save_xml_format(detections, labels_dir, image_file)

    def finalize_detections(self, num_train_images, num_val_images, num_total_images):
        expected_train_count = 0
        expected_val_count = 0 
        actual_total_images = self.actual_train_count + self.actual_val_count
        if actual_total_images == 0:
            self.log_update("No detections found in any image.")
        else:
            expected_train_count = int(actual_total_images * (1 - self.val_split))
            expected_val_count = actual_total_images - expected_train_count

            if (
                self.actual_train_count != expected_train_count
                or self.actual_val_count != expected_val_count
            ):
                self.adjust_split(expected_train_count, expected_val_count)

        self.train_detections = (
            self.train_detections if self.actual_train_count > 0 else []
        )
        self.val_detections = self.val_detections if self.actual_val_count > 0 else []

        if (
            self.actual_train_count == expected_train_count
            and self.actual_val_count == expected_val_count
        ):
            self.log_update(
                f"Split train/validation successful. Train: {self.actual_train_count}, Val: {self.actual_val_count}"
            )
        else:
            self.log_update(
                f"Split train/validation failed. Expected: train={expected_train_count}, val={expected_val_count}. Actual: train={self.actual_train_count}, val={self.actual_val_count}"
            )

        self.log_update(
            f"Results Finished. Processed {num_total_images} images: End Results is {self.actual_train_count} training images and {self.actual_val_count} validation images\n"
        )

    def adjust_split(self, expected_train_count, expected_val_count):
        train_images_dir = os.path.join(self.output_directory, "train", "images")
        val_images_dir = os.path.join(self.output_directory, "val", "images")
        train_detected_dir = os.path.join(self.output_directory, "train", "detected")
        train_labels_dir = os.path.join(self.output_directory, "train", "labels")
        val_detected_dir = os.path.join(self.output_directory, "val", "detected")
        val_labels_dir = os.path.join(self.output_directory, "val", "labels")

        while (
            self.actual_train_count != expected_train_count
            or self.actual_val_count != expected_val_count
        ):
            if self.actual_train_count < expected_train_count:
                val_images = os.listdir(val_images_dir)
                if not val_images:
                    break
                image = val_images.pop(0)
                src_image_path = os.path.join(val_images_dir, image)
                dst_image_path = os.path.join(train_images_dir, image)
                src_detected_path = os.path.join(val_detected_dir, image)
                dst_detected_path = os.path.join(train_detected_dir, image)
                src_label_path = os.path.join(
                    val_labels_dir, os.path.splitext(image)[0] + ".txt"
                )
                dst_label_path = os.path.join(
                    train_labels_dir, os.path.splitext(image)[0] + ".txt"
                )
                os.rename(src_image_path, dst_image_path)
                os.rename(src_detected_path, dst_detected_path)
                os.rename(src_label_path, dst_label_path)

                detections_to_move = [
                    det for det, img in self.val_detections if img == image
                ]
                self.train_detections.extend(
                    [(det, image) for det in detections_to_move]
                )
                self.val_detections = [
                    (det, img) for det, img in self.val_detections if img != image
                ]

                self.actual_train_count += 1
                self.actual_val_count -= 1
            elif self.actual_val_count < expected_val_count:
                train_images = os.listdir(train_images_dir)
                if not train_images:
                    break
                image = train_images.pop(0)
                src_image_path = os.path.join(train_images_dir, image)
                dst_image_path = os.path.join(val_images_dir, image)
                src_detected_path = os.path.join(train_detected_dir, image)
                dst_detected_path = os.path.join(val_detected_dir, image)
                src_label_path = os.path.join(
                    train_labels_dir, os.path.splitext(image)[0] + ".txt"
                )
                dst_label_path = os.path.join(
                    val_labels_dir, os.path.splitext(image)[0] + ".txt"
                )
                os.rename(src_image_path, dst_image_path)
                os.rename(src_detected_path, dst_detected_path)
                os.rename(src_label_path, dst_label_path)

                detections_to_move = [
                    det for det, img in self.train_detections if img == image
                ]
                self.val_detections.extend([(det, image) for det in detections_to_move])
                self.train_detections = [
                    (det, img) for det, img in self.train_detections if img != image
                ]

                self.actual_val_count += 1
                self.actual_train_count -= 1

    def get_detections(self):
        train_detections = [det for det, _ in self.train_detections]
        val_detections = [det for det, _ in self.val_detections]
        return (
            train_detections,
            val_detections,
            self.actual_train_count,
            self.actual_val_count,
        )

    def terminate(self):
        self._terminate_event.set()

    def log_update(self, log_message):
        if self.log_queue is not None:
            self.log_queue.put(log_message)

    def show_error(self, message):
        if self.show_error_queue is not None:
            self.show_error_queue.put(message)

    def output_image(self, output_img_path, image_file):
        if self.output_image_queue is not None:
            self.output_image_queue.put((output_img_path, image_file))

    def progress(self, value):
        if self.progress_queue is not None:
            self.progress_queue.put(value)



import queue
from queue import Queue
import customtkinter
import torch.cuda as cuda
from tkinter import messagebox, filedialog
from .workers.chart_worker import ChartWorker
from .workers.detection_worker import ObjectDetectionWorker


class MainWindow:
    def __init__(self, root):
        self.root = root
        self.root.title("ODLabel")
        self.root.geometry(
            "{}x{}".format(
                int(self.root.winfo_screenwidth() * 0.8),
                int(self.root.winfo_screenheight() * 0.8),
            )
        )
        self.root.protocol("WM_DELETE_WINDOW", self.on_closing)

        self.root.grid_rowconfigure(0, weight=1)
        self.root.grid_columnconfigure(0, weight=1)
        self.root.grid_columnconfigure(1, weight=90)

        # Left Panel
        self.left_frame = customtkinter.CTkFrame(self.root)
        self.left_frame.grid(row=0, column=0, padx=10, pady=10, sticky="nsew")

        # Configure grid weights for the left frame
        for i in range(26):
            self.left_frame.grid_rowconfigure(i, weight=1)
        self.left_frame.grid_columnconfigure(0, weight=1)

        # Model Path
        self.selected_model_path_label = customtkinter.CTkLabel(
            self.left_frame, text="", width=100
        )
        self.selected_model_path_label.grid(row=1, column=0, sticky="w")
        self.model_path_button = customtkinter.CTkButton(
            self.left_frame, text="Select YOLO-World Model", command=self.select_model
        )
        self.model_path_button.grid(row=0, column=0, sticky="s")

        # Images Folder
        self.selected_images_folder_label = customtkinter.CTkLabel(
            self.left_frame, text="", width=100
        )
        self.selected_images_folder_label.grid(row=3, column=0, sticky="w")
        self.images_folder_button = customtkinter.CTkButton(
            self.left_frame,
            text="Select Images Folder",
            command=self.select_images_folder,
        )
        self.images_folder_button.grid(row=2, column=0, sticky="s")

        # Output Directory
        self.selected_output_directory_label = customtkinter.CTkLabel(
            self.left_frame, text="", width=100
        )
        self.selected_output_directory_label.grid(row=5, column=0, sticky="w")
        self.output_directory_button = customtkinter.CTkButton(
            self.left_frame,
            text="Select Output Directory",
            command=self.select_output_directory,
        )
        self.output_directory_button.grid(row=4, column=0, sticky="s")

        # Classes
        self.classes_label = customtkinter.CTkLabel(
            self.left_frame, text="Object Categories:"
        )
        self.classes_label.grid(row=6, column=0, sticky="w")
        self.classes_entry = customtkinter.CTkEntry(
            self.left_frame,
            placeholder_text="Enter the classes to be detected, separated by comma",
        )
        self.classes_entry.grid(row=7, column=0, sticky="ew")
        self.classes_entry.bind("<KeyRelease>", self.update_classes)

        # SAHI Checkbox and Slice Size Input
        sahi_frame = customtkinter.CTkLabel(self.left_frame)
        sahi_frame.grid(row=8, column=0, pady=(0, 0), sticky="ew")

        self.sahi_var = customtkinter.BooleanVar()
        self.sahi_checkbox = customtkinter.CTkCheckBox(
            sahi_frame,
            variable=self.sahi_var,
            command=self.update_slice_size_input,
            height=20,
            text="Slicing Adaptive Inference (SAHI)",
        )
        self.sahi_checkbox.grid(row=0, column=0)

        # Slice Size Input
        self.slice_size_frame = customtkinter.CTkFrame(self.left_frame)
        self.slice_size_frame.grid(row=9, column=0, pady=(0, 10), sticky="s")
        self.slice_size_label = customtkinter.CTkLabel(
            self.slice_size_frame, text="Slice Size:"
        )
        self.slice_size_label.grid(row=0, column=0)
        self.slice_size_var = customtkinter.StringVar(value="512")
        self.slice_size_entry = customtkinter.CTkEntry(
            self.slice_size_frame, textvariable=self.slice_size_var, width=80
        )
        self.slice_size_entry.grid(row=0, column=1, padx=5)
        self.slice_size_entry.configure(state="disabled")

        # Device
        self.device_label = customtkinter.CTkLabel(self.left_frame, text="Device Type:")
        self.device_label.grid(row=10, column=0, sticky="sw")
        self.device_frame = customtkinter.CTkFrame(self.left_frame)
        self.device_frame.grid(row=10, column=0, sticky="s")
        self.device_var = customtkinter.StringVar(value="CPU")
        self.device_radiobutton_cpu = customtkinter.CTkRadioButton(
            self.device_frame, text="CPU", variable=self.device_var, value="CPU"
        )
        self.device_radiobutton_cpu.grid(row=0, column=0)
        self.device_radiobutton_gpu = customtkinter.CTkRadioButton(
            self.device_frame, text="GPU", variable=self.device_var, value="GPU"
        )
        self.device_radiobutton_gpu.grid(row=0, column=1)

        # Output Format
        self.output_format_label = customtkinter.CTkLabel(
            self.left_frame, text="Output Format:"
        )
        self.output_format_label.grid(row=12, column=0, sticky="sw")
        self.output_format_var = customtkinter.StringVar(value="YOLO")
        self.output_format_option_menu = customtkinter.CTkOptionMenu(
            self.left_frame,
            values=["YOLO", "COCO", "CSV", "XML"],
            variable=self.output_format_var,
        )
        self.output_format_option_menu.grid(row=12, column=0, sticky="s")

        # Train/Val Split
        self.split_label = customtkinter.CTkLabel(
            self.left_frame, text="Train/Val Split:"
        )
        self.split_label.grid(row=14, column=0, sticky="s")
        self.val_var = customtkinter.DoubleVar(value=0.8)
        self.val_slider = customtkinter.CTkSlider(
            self.left_frame,
            from_=0,
            to=1,
            number_of_steps=20,
            variable=self.val_var,
            command=self.update_val_label,
        )
        self.val_slider.grid(row=15, column=0, sticky="ew")
        self.val_label = customtkinter.CTkLabel(self.left_frame, text="0.80")
        self.val_label.grid(row=16, column=0)

        # Confidence
        self.conf_label = customtkinter.CTkLabel(
            self.left_frame, text="Confidence Level:"
        )
        self.conf_label.grid(row=17, column=0, sticky="s")
        self.conf_var = customtkinter.DoubleVar(value=0.2)
        self.conf_slider = customtkinter.CTkSlider(
            self.left_frame,
            from_=0,
            to=1,
            number_of_steps=100,
            variable=self.conf_var,
            command=self.update_conf_label,
        )
        self.conf_slider.grid(row=18, column=0, sticky="ew")
        self.conf_label = customtkinter.CTkLabel(self.left_frame, text="0.20")
        self.conf_label.grid(row=19, column=0)

        # Union over Intersection iou
        self.iou_label = customtkinter.CTkLabel(
            self.left_frame, text="Non Maximum Suppression:"
        )
        self.iou_label.grid(row=20, column=0, sticky="s")
        self.iou_var = customtkinter.DoubleVar(value=0.7)
        self.iou_slider = customtkinter.CTkSlider(
            self.left_frame,
            from_=0,
            to=1,
            number_of_steps=10,
            variable=self.iou_var,
            command=self.update_iou_label,
        )
        self.iou_slider.grid(row=21, column=0, sticky="ew")
        self.iou_label = customtkinter.CTkLabel(self.left_frame, text="0.70")
        self.iou_label.grid(row=22, column=0)

        # Run/Cancel Buttons
        self.run_cancel_frame = customtkinter.CTkFrame(self.left_frame)
        self.run_cancel_frame.grid(row=23, column=0, pady=10, sticky="s")
        self.run_button = customtkinter.CTkButton(
            self.run_cancel_frame, text="Start", command=self.run_code
        )
        self.run_button.grid(row=0, column=0, padx=5)
        self.cancel_button = customtkinter.CTkButton(
            self.run_cancel_frame, text="Cancel", command=self.cancel_code
        )
        self.cancel_button.grid(row=0, column=1, padx=5)
        self.cancel_button.configure(state="disabled")

        # Set the width of the output format option menu
        self.output_format_option_menu.configure(width=self.run_button.winfo_reqwidth())

        # Progress Bar
        self.progress_bar = customtkinter.CTkProgressBar(self.left_frame, width=400)
        self.progress_bar.grid(row=24, column=0, sticky="ew")
        self.progress_bar.set(0)

        # Log Text
        self.log_text = customtkinter.CTkTextbox(self.left_frame, width=400, height=200)
        self.log_text.grid(row=25, column=0, sticky="nsew")
        self.log_text.configure(state="disabled")

        # chart frame
        self.charts_frame = customtkinter.CTkFrame(self.root)
        self.charts_frame.grid(row=0, column=1, padx=10, pady=10, sticky="nsew")

        self.tab_control = customtkinter.CTkTabview(self.charts_frame)
        self.tab_control.grid(row=0, column=0, sticky="nsew")

        self.input_tab = self.tab_control.add("Input")
        self.output_tab = self.tab_control.add("Output")

        self.input_chart_frame = customtkinter.CTkFrame(self.input_tab)
        self.input_chart_frame.grid(row=0, column=0, sticky="nsew")

        self.output_tab_control = customtkinter.CTkTabview(self.output_tab)
        self.output_tab_control.grid(row=0, column=0, sticky="nsew")

        self.train_tab = self.output_tab_control.add("Train")
        self.val_tab = self.output_tab_control.add("Val")

        self.train_chart_frame = customtkinter.CTkFrame(self.train_tab)
        self.train_chart_frame.grid(row=0, column=0, sticky="nsew")

        self.val_chart_frame = customtkinter.CTkFrame(self.val_tab)
        self.val_chart_frame.grid(row=0, column=0, sticky="nsew")

        self.charts_frame.grid_rowconfigure(0, weight=1)
        self.charts_frame.grid_columnconfigure(0, weight=1)

        self.input_tab.grid_rowconfigure(0, weight=1)
        self.input_tab.grid_columnconfigure(0, weight=1)

        self.output_tab.grid_rowconfigure(0, weight=1)
        self.output_tab.grid_columnconfigure(0, weight=1)

        self.train_tab.grid_rowconfigure(0, weight=1)
        self.train_tab.grid_columnconfigure(0, weight=1)

        self.val_tab.grid_rowconfigure(0, weight=1)
        self.val_tab.grid_columnconfigure(0, weight=1)

        self.chart_queue = Queue()
        self.worker_thread = None
        self.classes = []
        self.input_figures = None
        self.train_figures = None
        self.val_figures = None
        self.input_chart_worker = None
        self.output_chart_worker = None
        self.show_error_queue = None
        self.log_queue = None
        self.output_image_queue = None
        self.progress_queue = None
        self.train_figures_processed = False
        self.val_figures_processed = False

        # Chart Colors
        self.chart_colors = {
            "images": "#1f77b4",
            "resolution": "#1f77b4",
            "quality": "#1f77b4",
            "format": "#1f77b4",
        }

    def update_classes(self, event):
        text = self.classes_entry.get()
        self.classes = [c.strip() for c in text.split(",")]

    def update_slice_size_input(self):
        if self.sahi_var.get():
            self.slice_size_entry.configure(state="normal")
            self.slice_size_label.configure(state="normal")
        else:
            self.slice_size_entry.configure(state="disabled")
            self.slice_size_label.configure(state="disabled")

    def validate_slice_size(self, value):
        if value.isdigit() and int(value) % 32 == 0:
            return True
        return False

    def select_model(self):
        file_path = filedialog.askopenfilename(filetypes=[("YOLO-WORLD Model", "*.pt")])
        if file_path:
            self.selected_model_path_label.configure(text=file_path)
            self.check_run_button_state()

    def select_images_folder(self):
        folder_path = filedialog.askdirectory()
        if folder_path:
            self.selected_images_folder_label.configure(text=folder_path)
            self.check_run_button_state()

            try:
                self.clear_input_chart()
                self.log_text.configure(state="normal")
                self.log_text.insert(customtkinter.END, "Processing input charts...\n")
                self.log_text.configure(state="disabled")
                self.input_chart_worker = ChartWorker(folder_path, self.chart_colors)
                self.input_chart_worker.queue = self.chart_queue
                self.input_chart_worker.start()
                self.monitor_input_chart_worker()
            except Exception as e:
                self.show_error_message(str(e))

    def select_output_directory(self):
        folder_path = filedialog.askdirectory()
        if folder_path:
            self.selected_output_directory_label.configure(text=folder_path)
            self.check_run_button_state()

    def clear_input_chart(self):
        if self.input_chart_worker:
            self.input_chart_worker.close_figures()
        for widget in self.input_chart_frame.winfo_children():
            widget.destroy()

    def update_val_label(self, value):
        self.val_label.configure(text=f"{float(value):.2f}")

    def update_conf_label(self, value):
        self.conf_label.configure(text=f"{float(value):.2f}")

    def update_iou_label(self, value):
        self.iou_label.configure(text=f"{float(value):.2f}")

    def check_run_button_state(self):
        model_path = self.selected_model_path_label.cget("text")
        images_folder = self.selected_images_folder_label.cget("text")
        output_directory = self.selected_output_directory_label.cget("text")
        if model_path and images_folder and output_directory:
            self.run_button.configure(state="normal")
        else:
            self.run_button.configure(state="disabled")

    def run_code(self):
        model_path = self.selected_model_path_label.cget("text")
        images_folder = self.selected_images_folder_label.cget("text")
        output_directory = self.selected_output_directory_label.cget("text")
        use_sahi = self.sahi_var.get()
        slice_size = int(self.slice_size_var.get())
        train_split = self.val_var.get()
        val_split = 1 - train_split
        conf_threshold = self.conf_var.get()
        iou_threshold = self.iou_var.get()
        output_format = self.output_format_var.get()
        device = "cpu"

        if not self.classes:
            self.show_error_message("Please enter class names.")
            return
        
        if self.device_var.get() == "GPU":
            gpu_available = cuda.is_available()
            if gpu_available:
                device = "cuda"
                gpu_name = cuda.get_device_name(0)
                self.log_text.configure(state="normal")
                self.log_text.insert(customtkinter.END, f"Using GPU: {gpu_name}\n")
                self.log_text.configure(state="disabled")
            else:
                device = "cpu"
                self.log_text.configure(state="normal")
                self.log_text.insert(customtkinter.END, "No GPU found. Using CPU instead.\n")
                self.log_text.configure(state="disabled")
                self.device_radiobutton_cpu.invoke()       
        else:
            self.log_text.configure(state="normal")
            self.log_text.insert(customtkinter.END, "Using CPU.\n")
            self.log_text.configure(state="disabled") 

        try:
            self.clear_output()
            self.worker_thread = ObjectDetectionWorker(
                conf_threshold,
                model_path,
                images_folder,
                output_directory,
                val_split,
                output_format,
                self.classes,
                device,
                use_sahi,
                slice_size,
                iou_threshold,
            )
            self.worker_thread.log_queue = queue.Queue()
            self.worker_thread.progress_queue = queue.Queue()
            self.worker_thread.show_error_queue = queue.Queue()

            self.progress_bar.set(0)
            self.worker_thread.start()

            self.monitor_worker_thread()

            self.log_text.configure(state="normal")
            self.log_text.insert(customtkinter.END, "Process Started.\n")
            self.log_text.configure(state="disabled")

            self.update_ui_state("disabled")

        except Exception as e:
            self.show_error_message(str(e))

    def monitor_worker_thread(self):
        if self.worker_thread is None:
            return

        if self.worker_thread.is_alive():
            if (
                self.worker_thread.progress_queue is not None
                and not self.worker_thread.progress_queue.empty()
            ):
                progress_value = self.worker_thread.progress_queue.get()
                self.update_progress(progress_value)

            if (
                self.worker_thread.show_error_queue is not None
                and not self.worker_thread.show_error_queue.empty()
            ):
                error_message = self.worker_thread.show_error_queue.get()
                self.show_error_message(error_message)

            if (
                self.worker_thread.log_queue is not None
                and not self.worker_thread.log_queue.empty()
            ):
                log_message = self.worker_thread.log_queue.get()
                self.update_log(log_message)

            self.root.after(100, self.monitor_worker_thread)
        else:
            self.on_worker_finished()

    def monitor_input_chart_worker(self):
        if self.input_chart_worker is None:
            return

        if not self.input_chart_worker.is_alive():
            while not self.chart_queue.empty():
                chart_data = self.chart_queue.get()
                chart_type, data = chart_data
                if chart_type == "input_figures":
                    self.process_input_figures(data)
                    self.on_input_chart_worker_finished()
                elif chart_type == "error":
                    self.log_text.configure(state="normal")
                    self.log_text.insert(
                        customtkinter.END, f"Error generating input charts: {data}\n"
                    )
                    self.log_text.configure(state="disabled")
                    self.show_error_message("No images found in the selected folder.")

        else:
            self.root.after(100, self.monitor_input_chart_worker)

    def monitor_output_chart_worker(self):
        if self.output_chart_worker is None:
            return

        self.train_figures_processed = False
        self.val_figures_processed = False

        if not self.output_chart_worker.is_alive():
            while not self.chart_queue.empty():
                chart_data = self.chart_queue.get()
                chart_type, figures = chart_data
                if chart_type == "train_figures":
                    self.process_train_figures(figures)
                    self.train_figures_processed = True
                elif chart_type == "val_figures":
                    self.process_val_figures(figures)
                    self.val_figures_processed = True

            self.on_output_chart_worker_finished()
        else:
            self.root.after(100, self.monitor_output_chart_worker)

    def clear_output(self):
        for widget in self.train_chart_frame.winfo_children():
            widget.destroy()
        for widget in self.val_chart_frame.winfo_children():
            widget.destroy()

    def cancel_code(self):
        if self.worker_thread and self.worker_thread.is_alive():
            self.worker_thread.terminate()
            self.worker_thread.join()
            self.worker_thread = None
            self.progress_bar.set(0)

            self.log_text.configure(state="normal")
            self.log_text.insert(customtkinter.END, "Process Canceled.\n")
            self.log_text.configure(state="disabled")

            self.update_ui_state("normal")

        self.on_worker_finished()

    def process_input_figures(self, figures):
        self.clear_input_chart()
        if figures and self.input_chart_worker is not None:
            self.input_chart_worker.display_chart_in_frame(
                figures[0], self.input_chart_frame, row=0, col=0
            )
            self.input_chart_worker.display_chart_in_frame(
                figures[1], self.input_chart_frame, row=0, col=1
            )
            self.input_chart_worker.display_chart_in_frame(
                figures[2], self.input_chart_frame, row=1, col=0
            )
            self.input_chart_worker.display_chart_in_frame(
                figures[3], self.input_chart_frame, row=1, col=1
            )

    def process_train_figures(self, figures):
        self.clear_output_charts(self.train_chart_frame)
        if figures and self.output_chart_worker is not None:
            self.output_chart_worker.display_chart_in_frame(
                figures[0], self.train_chart_frame, row=0, col=0
            )
            self.output_chart_worker.display_chart_in_frame(
                figures[1], self.train_chart_frame, row=1, col=1
            )
            self.output_chart_worker.display_chart_in_frame(
                figures[2], self.train_chart_frame, row=0, col=1
            )
            self.output_chart_worker.display_chart_in_frame(
                figures[3], self.train_chart_frame, row=1, col=0
            )

    def process_val_figures(self, figures):
        self.clear_output_charts(self.val_chart_frame)
        if figures and self.output_chart_worker is not None:
            self.output_chart_worker.display_chart_in_frame(
                figures[0], self.val_chart_frame, row=0, col=0
            )
            self.output_chart_worker.display_chart_in_frame(
                figures[1], self.val_chart_frame, row=1, col=1
            )
            self.output_chart_worker.display_chart_in_frame(
                figures[2], self.val_chart_frame, row=0, col=1
            )
            self.output_chart_worker.display_chart_in_frame(
                figures[3], self.val_chart_frame, row=1, col=0
            )

    def on_input_chart_worker_finished(self):
        self.log_text.configure(state="normal")
        self.log_text.insert(
            customtkinter.END, "Input charts processed successfully.\n"
        )
        self.log_text.configure(state="disabled")

        self.input_chart_worker = None

    def on_output_chart_worker_finished(self):
        if self.train_figures_processed and self.val_figures_processed:
            self.log_text.configure(state="normal")
            self.log_text.insert(
                customtkinter.END, "Output charts processed successfully.\n"
            )
            self.log_text.configure(state="disabled")
        elif self.train_figures_processed and not self.val_figures_processed:
            if self.val_var.get() > 0:
                self.log_text.configure(state="normal")
                self.log_text.insert(
                    customtkinter.END, "Training charts processed successfully.\n"
                )
                self.log_text.insert(
                    customtkinter.END, "No detections found in validation images.\n"
                )
                self.log_text.configure(state="disabled")
            else:
                self.log_text.configure(state="normal")
                self.log_text.insert(
                    customtkinter.END, "Training charts processed successfully.\n"
                )
                self.log_text.configure(state="disabled")
        elif not self.train_figures_processed and self.val_figures_processed:
            if self.val_var.get() < 1:
                self.log_text.configure(state="normal")
                self.log_text.insert(
                    customtkinter.END, "No detections found in training images.\n"
                )
                self.log_text.insert(
                    customtkinter.END, "Validation charts processed successfully.\n"
                )
                self.log_text.configure(state="disabled")
            else:
                self.log_text.configure(state="normal")
                self.log_text.insert(
                    customtkinter.END, "Validation charts processed successfully.\n"
                )
                self.log_text.configure(state="disabled")
        elif not self.train_figures_processed and not self.val_figures_processed:
            if self.val_var.get() < 1:
                self.log_text.configure(state="normal")
                self.log_text.insert(
                    customtkinter.END, "No detections found in training images.\n"
                )
                self.log_text.configure(state="disabled")
            if self.val_var.get() > 0:
                self.log_text.configure(state="normal")
                self.log_text.insert(
                    customtkinter.END, "No detections found in validation images.\n"
                )
                self.log_text.configure(state="disabled")

        self.output_chart_worker = None

    def on_worker_finished(self):
        if self.worker_thread is not None:
            train_detections, val_detections, actual_train_count, actual_val_count = (
                self.worker_thread.get_detections()
            )

            if self.worker_thread.log_queue is not None:
                while not self.worker_thread.log_queue.empty():
                    log_message = self.worker_thread.log_queue.get()
                    self.update_log(log_message)

            if self.worker_thread.show_error_queue is not None:
                while not self.worker_thread.show_error_queue.empty():
                    error_message = self.worker_thread.show_error_queue.get()
                    self.show_error_message(error_message)

        else:
            train_detections, val_detections, actual_train_count, actual_val_count = (
                [],
                [],
                0,
                0,
            )

        self.worker_thread = None
        self.progress_bar.set(100)

        self.log_text.configure(state="normal")
        self.log_text.insert(customtkinter.END, "Processing output charts...\n")
        self.log_text.configure(state="disabled")

        self.update_ui_state("normal")

        if actual_train_count == 0 and actual_val_count == 0:
            self.log_text.configure(state="normal")
            self.log_text.insert(
                customtkinter.END, "No detections found in any image.\n"
            )
            self.log_text.configure(state="disabled")
            return

        try:
            self.output_chart_worker = ChartWorker(
                self.selected_images_folder_label.cget("text"),
                self.chart_colors,
                train_detections,
                val_detections,
                actual_train_count,
                actual_val_count,
            )
            self.output_chart_worker.queue = self.chart_queue
            self.output_chart_worker.start()
            self.monitor_output_chart_worker()
        except Exception as e:
            self.show_error_message(str(e))

    def clear_output_charts(self, frame):
        for widget in frame.winfo_children():
            widget.destroy()

    def display_chart(self, fig, frame, row, col):
        if self.input_chart_worker is not None:
            self.input_chart_worker.display_chart_in_frame(fig, frame, row, col)

    def update_log(self, log_message):
        self.log_text.configure(state="normal")
        self.log_text.insert(customtkinter.END, log_message + "\n")
        self.log_text.see(customtkinter.END)
        self.log_text.configure(state="disabled")

    def show_error_message(self, message):
        messagebox.showerror("Error", message)

    def update_progress(self, progress_value):
        self.progress_bar.set(progress_value / 100)

    def update_ui_state(self, state):
        self.run_button.configure(state=state)
        self.cancel_button.configure(
            state="normal" if state == "disabled" else "disabled"
        )
        self.model_path_button.configure(state=state)
        self.selected_model_path_label.configure(state=state)
        self.images_folder_button.configure(state=state)
        self.selected_images_folder_label.configure(state=state)
        self.output_directory_button.configure(state=state)
        self.selected_output_directory_label.configure(state=state)
        self.sahi_checkbox.configure(state=state)
        self.classes_entry.configure(state=state)
        self.classes_label.configure(state=state)
        self.slice_size_label.configure(state=state)
        self.slice_size_entry.configure(state=state)
        self.device_label.configure(state=state)
        self.device_radiobutton_cpu.configure(state=state)
        self.device_radiobutton_gpu.configure(state=state)
        self.output_format_option_menu.configure(state=state)
        self.conf_slider.configure(state=state)
        self.val_slider.configure(state=state)
        self.iou_slider.configure(state=state)

    def on_closing(self):
        if self.input_chart_worker:
            self.input_chart_worker.close_figures()
        if self.output_chart_worker:
            self.output_chart_worker.close_figures()
        self.root.quit()
        self.root.destroy()



import os
import cv2 as cv
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg


def input_chart(images_folder, chart_colors):
    image_files = [
        f
        for f in os.listdir(images_folder)
        if f.endswith((".jpg", ".jpeg", ".png", "jfif"))
    ]
    resolutions = []
    blurred_images = 0
    grayed_images = 0
    black_white_images = 0
    corrupted_images = 0
    color_spaces = {}

    for image_file in image_files:
        img_path = os.path.join(images_folder, image_file)
        try:
            img = cv.imread(img_path)
            if img is not None:
                height, width = img.shape[:2]
                resolutions.append((width, height))

                # Check for blurred images
                if cv.Laplacian(img, cv.CV_64F).var() < 100:
                    blurred_images += 1

                # Check for grayscale images
                if len(img.shape) < 3 or img.shape[2] == 1:
                    grayed_images += 1

                # Check for black and white images
                if (
                    len(img.shape) == 3
                    and np.all(img[..., 0] == img[..., 1])
                    and np.all(img[..., 1] == img[..., 2])
                ):
                    black_white_images += 1

                # Check color space
                color_space = tuple(img.mean(axis=(0, 1)) / 255.0)
                color_spaces[color_space] = color_spaces.get(color_space, 0) + 1
        except Exception:
            corrupted_images += 1

    # Plot number of images and image format distribution
    fig_format = Figure(
        figsize=(4, 3),
        dpi=100,
        facecolor="#1c1c1c",
        edgecolor="#1c1c1c",
        tight_layout=True,
    )
    ax_format = fig_format.add_subplot(1, 1, 1)
    formats = [os.path.splitext(f)[1].lower() for f in image_files]
    format_counts = {}
    for fmt in formats:
        format_counts[fmt] = format_counts.get(fmt, 0) + 1

    _ = ax_format.bar(
        ["Images"], [len(image_files)], color=chart_colors["images"], width=0.4
    )
    _ = ax_format.bar(
        list(format_counts.keys()),
        list(format_counts.values()),
        color=chart_colors["format"],
        width=0.4,
    )
    ax_format.set_ylabel("Instances", color="white")
    ax_format.set_title("Instances and Format Distribution", color="white")
    ax_format.tick_params(axis="x", colors="white", rotation=0)
    ax_format.tick_params(axis="y", colors="white")
    fig_format.set_facecolor("#1c1c1c")  # dark gray

    # Plot image resolution distribution
    fig_resolution = Figure(
        figsize=(4, 3),
        dpi=100,
        facecolor="#1c1c1c",
        edgecolor="#1c1c1c",
        tight_layout=True,
    )
    ax_resolution = fig_resolution.add_subplot(1, 1, 1)
    widths = [res[0] for res in resolutions]
    min_width = min(widths)
    # max_width = max(widths)
    bins = np.arange(min_width, 4000, 600)

    ax_resolution.hist(widths, bins=bins, color=chart_colors["resolution"])
    ax_resolution.set_xlabel("Resolution", color="white")
    ax_resolution.set_ylabel("Instances", color="white")
    ax_resolution.set_title("Image Resolution Distribution", color="white")
    ax_resolution.tick_params(axis="x", colors="white")
    ax_resolution.tick_params(axis="y", colors="white")
    ax_resolution.ticklabel_format(axis="x", style="plain")
    ax_resolution.set_xticks(bins)
    ax_resolution.set_xticklabels(list(map(str, bins)), rotation=0, ha="center")
    fig_resolution.set_facecolor("#1c1c1c")  # dark gray

    # Plot image quality information
    fig_quality = Figure(
        figsize=(4, 3),
        dpi=100,
        facecolor="#1c1c1c",
        edgecolor="#1c1c1c",
        tight_layout=True,
    )
    ax_quality = fig_quality.add_subplot(1, 1, 1)
    quality_data = [blurred_images, grayed_images, black_white_images, corrupted_images]
    ax_quality.bar(
        ["Blurred", "Grayscale", "B&W", "Corrupted"],
        quality_data,
        color=chart_colors["quality"],
    )
    ax_quality.set_ylabel("Instances", color="white")
    ax_quality.set_title("Image Quality", color="white")
    ax_quality.tick_params(axis="x", colors="white")
    ax_quality.tick_params(axis="y", colors="white", labelleft=True, labelright=False)
    ax_quality.yaxis.set_major_locator(plt.MaxNLocator(integer=True))
    fig_quality.set_facecolor("#1c1c1c")  # dark gray

    # Plot color space distribution
    fig_colorspace = Figure(
        figsize=(4, 3),
        dpi=100,
        facecolor="#1c1c1c",
        edgecolor="#1c1c1c",
        tight_layout=True,
    )
    ax_colorspace = fig_colorspace.add_subplot(1, 1, 1, projection="3d")
    color_spaces_keys = list(color_spaces.keys())
    x = [color[0] for color in color_spaces_keys]
    y = [color[1] for color in color_spaces_keys]
    z = [color[2] for color in color_spaces_keys]
    sizes = [size * 50 for size in list(color_spaces.values())]

    ax_colorspace.scatter(x, y, z, c=color_spaces_keys, s=sizes)

    ax_colorspace.set_xlabel("Red", color="white")
    ax_colorspace.set_ylabel("Green", color="white")
    ax_colorspace.set_zlabel("Blue", color="white")
    ax_colorspace.set_title("Color Space Distribution", color="white")
    ax_colorspace.tick_params(axis="x", colors="white")
    ax_colorspace.tick_params(axis="y", colors="white")
    ax_colorspace.tick_params(axis="z", colors="white")
    ax_colorspace.set_facecolor((0, 0, 0, 0))
    fig_colorspace.set_facecolor("#1c1c1c")  # dark gray

    return fig_format, fig_resolution, fig_quality, fig_colorspace, image_files


def output_charts(detections, num_images):
    object_counts = {}
    confidences = []
    labels_sizes = {}
    bboxes = []
    class_names = []
    heatmap_coords = {}
    labelism_counts = {}

    for _, x, y, w, h, class_name, confidence, img_width, img_height in detections:
        object_counts[class_name] = object_counts.get(class_name, 0) + 1
        confidences.append(confidence)
        labels_sizes[class_name] = labels_sizes.get(class_name, 0) + 1

        # Normalizing bounding box coordinates to [0, 1] range
        normalized_x = x / img_width
        normalized_y = y / img_height
        normalized_w = w / img_width
        normalized_h = h / img_height

        # Adjusting width and x-coordinate to account for the bounding box origin
        normalized_x -= normalized_w / 2
        normalized_w = max(0, min(1 - normalized_x, normalized_w))
        normalized_y -= normalized_h / 2
        normalized_h = max(0, min(1 - normalized_y, normalized_h))

        # Flipping the y-coordinate to match the chart's coordinate system
        normalized_y = 1 - normalized_y - normalized_h

        bboxes.append((normalized_x, normalized_y, normalized_w, normalized_h))

        # Update heatmap and labelism counts
        heatmap_center_x = normalized_x + normalized_w / 2
        heatmap_center_y = normalized_y + normalized_h / 2
        heatmap_coords[(heatmap_center_x, heatmap_center_y)] = (
            heatmap_coords.get((heatmap_center_x, heatmap_center_y), 0) + 1
        )

        labelism_counts[(normalized_x, normalized_y, normalized_w, normalized_h)] = (
            labelism_counts.get(
                (normalized_x, normalized_y, normalized_w, normalized_h), 0
            )
            + 1
        )
        class_names.append(class_name)

    # Count of Detected Objects
    fig_count = Figure(figsize=(4, 3), dpi=100)
    ax_count = fig_count.add_subplot(1, 1, 1)

    if object_counts:
        x = np.arange(len(object_counts) + 1)
        width = 0.4
        ax_count.bar(x[0], num_images, width, color="blue", label="Images")
        ax_count.bar(
            x[1:], list(object_counts.values()), width, color="orange", label="Classes"
        )
        ax_count.set_xticks(x)
        ax_count.set_xticklabels(
            ["Images"] + list(object_counts.keys()), rotation=0, ha="center"
        )
    else:
        ax_count.bar(0, num_images, 0.4, color="blue", label="Images")
        ax_count.set_xticks([0])
        ax_count.set_xticklabels(["Images"], rotation=0, ha="center")

    ax_count.set_ylabel("Instance")
    ax_count.set_title("Instances of Detected Objects by Class", y=1.05)
    ax_count.legend()

    # Detection Confidence Histogram
    fig_conf = Figure(figsize=(4, 3), dpi=100)
    ax_conf = fig_conf.add_subplot(1, 1, 1)
    ax_conf.hist(confidences, bins=20, range=(0, 1))
    ax_conf.set_xlim(0, 1)
    ax_conf.set_xlabel("Confidence")
    ax_conf.set_ylabel("Frequency")
    ax_conf.set_title("Detection Confidence Histogram", y=1.05)

    # Spatial Distribution of Objects
    fig_labels = Figure(figsize=(4, 3), dpi=100)
    ax_labels = fig_labels.add_subplot(1, 1, 1)

    max_count = max(labelism_counts.values())
    cmap = plt.get_cmap("Reds")

    for (x, y, w, h), count in labelism_counts.items():
        rect = Rectangle(
            (x, y),
            w,
            h,
            linewidth=0.7,
            edgecolor=cmap(count / max_count),
            facecolor="none",
        )
        ax_labels.add_patch(rect)

    ax_labels.set_xlim(0, 1)
    ax_labels.set_ylim(0, 1)
    ax_labels.set_title("Spatial Distribution of Objects", y=1.05)
    ax_labels.set_xlabel("X")
    ax_labels.set_ylabel("Y")

    # Heatmap of Detection
    fig_heatmap = Figure(figsize=(4, 3), dpi=100)
    ax_heatmap = fig_heatmap.add_subplot(1, 1, 1)

    max_count = max(heatmap_coords.values())
    cmap = plt.get_cmap("Blues")

    for (x, y), count in heatmap_coords.items():
        color = cmap(count / max_count)
        ax_heatmap.plot(
            x, y, marker="o", markersize=1, color=color, markeredgecolor=color
        )

    ax_heatmap.set_xlim(0, 1)
    ax_heatmap.set_ylim(0, 1)
    ax_heatmap.set_title("Heatmap of Detection")
    ax_heatmap.set_xlabel("X")
    ax_heatmap.set_ylabel("Y")

    return fig_count, fig_conf, fig_labels, fig_heatmap


def display_chart(fig, frame, row, col):
    if isinstance(fig, tuple):
        fig, _ = fig
    else:
        _ = fig.get_axes()[0]
    canvas = FigureCanvasTkAgg(fig, master=frame)
    canvas.get_tk_widget().grid(row=row, column=col, sticky="nsew")
    frame.grid_rowconfigure(row, weight=1)
    frame.grid_columnconfigure(col, weight=1)
    return canvas


def close_matplotlib_figures():
    plt.close("all")



from setuptools import setup, find_packages

setup(
    name="odlabel",
    version="0.7.26.8",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "customtkinter",
        "ultralytics",
        "matplotlib",
        "numpy",
        "opencv-python",
        "torch>=1.7.0",
        "torchvision>=0.8.0",
        "clip-for-odlabel",
    ],
    entry_points={
        "console_scripts": [
            "odlabel=app.main:launch_GUI",
        ],
    },
    author="Ziad-Algrafi",
    author_email="ZiadAlgrafi@gmail.com",
    description="A tool for object detection, labeling and visualization using PyTorch",
    long_description=open("../README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Ziad-Algrafi/odlabel",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    python_requires=">=3.9",
    setup_requires=["setuptools>=38.6.0", "wheel"],
)


def infer(model, img, conf_threshold, log_update, image_file, iou_threshold):
    annotated_frame = img.copy()
    img_height, img_width = img.shape[:2]
    results = model(
        img, conf=conf_threshold, imgsz=640, iou=iou_threshold, verbose=False
    )
    detections = []
    names = model.names
    if results[0].boxes is not None:
        for box, class_id, score in zip(
            results[0].boxes.xywhn.cpu(),
            results[0].boxes.cls.int().cpu().tolist(),
            results[0].boxes.conf.cpu(),
        ):
            x, y, w, h = box
            cls_id = class_id
            confidence = float(score)
            class_name = names[cls_id]
            x_scaled = x * img_width
            y_scaled = y * img_height
            w_scaled = w * img_width
            h_scaled = h * img_height
            detections.append(
                (
                    cls_id,
                    x_scaled,
                    y_scaled,
                    w_scaled,
                    h_scaled,
                    class_name,
                    confidence,
                    img_width,
                    img_height,
                )
            )

    log_update_with_detections(log_update, detections, image_file)

    return detections, annotated_frame


def slice_and_infer(
    model,
    img,
    conf_threshold,
    log_update,
    image_file,
    slice_size,
    iou_threshold,
    terminate,
):
    annotated_frame = img.copy()
    img_height, img_width = img.shape[:2]
    slices = []
    num_slices = 0
    for y in range(0, img_height, slice_size // 2):
        for x in range(0, img_width, slice_size // 2):
            slice_y = min(y + slice_size, img_height)
            slice_x = min(x + slice_size, img_width)
            slice = img[y:slice_y, x:slice_x]
            slices.append((slice, x, y, slice_x - x, slice_y - y))
            num_slices += 1

    log_update(f"Number of slices: {num_slices}")
    detections = []
    names = model.names
    for i, (slice, x_offset, y_offset, slice_width, slice_height) in enumerate(
        slices, start=1
    ):
        results = model(
            slice,
            conf=conf_threshold,
            imgsz=slice_size,
            iou=iou_threshold,
            verbose=False,
        )[0]
        if results.boxes is not None:
            log_update(f"Slice {i} has detections.")
            for box in results.boxes:
                x, y, w, h = box.xywhn[0].tolist()
                cls_id = box.cls[0].int().item()
                confidence = box.conf[0].cpu()
                class_name = names[cls_id]
                x_pixel = x * slice_width
                y_pixel = y * slice_height
                w_pixel = w * slice_width
                h_pixel = h * slice_height
                x_global = x_pixel + x_offset
                y_global = y_pixel + y_offset
                x_scaled = x_global
                y_scaled = y_global
                w_scaled = w_pixel
                h_scaled = h_pixel
                detections.append(
                    (
                        cls_id,
                        x_scaled,
                        y_scaled,
                        w_scaled,
                        h_scaled,
                        class_name,
                        confidence,
                        img_width,
                        img_height,
                    )
                )
        else:
            log_update(f"Slice {i} has no detections.")

    detections = apply_nms(detections, iou_threshold)
    log_update_with_detections(log_update, detections, image_file)
    return detections, annotated_frame


def apply_nms(detections, iou_threshold):
    boxes = []
    for det in detections:
        (
            cls_id,
            x_scaled,
            y_scaled,
            w_scaled,
            h_scaled,
            class_name,
            confidence,
            img_width,
            img_height,
        ) = det
        x1 = x_scaled
        y1 = y_scaled
        x2 = x_scaled + w_scaled
        y2 = y_scaled + h_scaled
        boxes.append([x1, y1, x2, y2, cls_id, confidence])

    sorted_boxes = sorted(boxes, key=lambda x: x[5], reverse=True)
    selected_boxes = []
    while len(sorted_boxes) > 0:
        selected_boxes.append(sorted_boxes[0])
        remaining_boxes = []
        for box in sorted_boxes[1:]:
            x1_a, y1_a, x2_a, y2_a, *_ = selected_boxes[-1]
            x1_b, y1_b, x2_b, y2_b, *_ = box
            intersection = max(0, min(x2_a, x2_b) - max(x1_a, x1_b)) * max(
                0, min(y2_a, y2_b) - max(y1_a, y1_b)
            )
            union = (
                (x2_a - x1_a) * (y2_a - y1_a)
                + (x2_b - x1_b) * (y2_b - y1_b)
                - intersection
            )
            if union == 0:
                iou = 0
            else:
                iou = intersection / union
            if iou < iou_threshold:
                remaining_boxes.append(box)
        sorted_boxes = remaining_boxes

    nms_detections = []
    for box in selected_boxes:
        x1, y1, x2, y2, cls_id, confidence = box
        w_scaled = x2 - x1
        h_scaled = y2 - y1
        x_scaled = x1
        y_scaled = y1
        nms_detections.append(
            (
                cls_id,
                x_scaled,
                y_scaled,
                w_scaled,
                h_scaled,
                class_name,
                confidence,
                img_width,
                img_height,
            )
        )

    return nms_detections


def log_update_with_detections(log_update, detections, image_file):
    valid_detections = [det for det in detections if det[5] != ""]
    if not valid_detections:
        log_update(f"No Results for image {image_file}")
    else:
        class_counts = {}
        for _, _, _, _, _, class_name, _, _, _ in valid_detections:
            class_counts[class_name] = class_counts.get(class_name, 0) + 1

        log_message = f"Results for {image_file}:"
        for class_name, count in class_counts.items():
            plural_class_name = class_name + "s" if count > 1 else class_name
            log_message += f" {count} {plural_class_name}"
        log_update(log_message)
